const {replace} = '';
const ca = /[&<>'"]/g;
const esca = {
  '&': '&amp;',
  '<': '&lt;',
  '>': '&gt;',
  "'": '&#39;',
  '"': '&quot;'
};
const pe = m => esca[m];
const escape = es => replace.call(es, ca, pe);

const FluentFlags = (cb, flags = {})=>new Proxy(cb, {
        get (_, prop, receiver) {
            flags[prop] = true;
            return receiver;
        },
        apply (_, self, args) {
            const f = {
                ...flags
            };
            flags = {}
            ;
            return cb.call(self, f).apply(self, args);
        }
    })
;

function onEvent(el, type, listener, options) {
    el.addEventListener(type, listener, options);
    return ()=>off(el, type, listener, options)
    ;
}
 const on = ()=>FluentFlags((flags)=>(el, type, fn, options = {})=>onEvent(el, type, fn, {
                ...flags.active ? {
                    passive: false
                } : null,
                ...flags,
                ...options
            })
    )
;
 const off = (el, type, listener, options)=>el.removeEventListener(type, listener, options)
;

const accessors = (target, source, fn)=>Object.defineProperties(target, Object.fromEntries(Object.entries(source).map(([key, value])=>{
        const next = fn(key, value);
        const prev = Object.getOwnPropertyDescriptor(target, key);
        if (prev && prev.get && prev.set) {
            const { get , set  } = next;
            next.get = ()=>prev.get?.() ?? get()
            ;
            next.set = (v)=>{
                prev.set?.(v);
                set(prev.get?.() ?? v);
            };
        }
        return [
            key,
            {
                configurable: true,
                enumerable: true,
                ...next
            }
        ];
    })))
;

const Types = new Map([
    [
        String,
        (x)=>x.toString()
    ],
    [
        Number,
        (x)=>parseFloat(x)
    ],
    [
        Boolean,
        (x)=>x = x === false ? false : x != null
    ],
]);
const applyProps = (self, data)=>{
    for (const [key1, value1] of Object.entries(data))if ([
        String,
        Number,
        Boolean
    ].includes(value1)) data[key1] = void 0;
    accessors(self, data, (key)=>{
        const attr = key.toLowerCase();
        return {
            get: ()=>data[key]
            ,
            set (value) {
                self.setAttribute(attr, value);
            }
        };
    });
    return data;
};
const attrs = (attrs1)=>(superclass)=>{
        const types = new Map();
        const map = new Map();
        for (const [key2, value] of Object.entries(new attrs1())){
            const type = value == null ? Types.get(String) : Types.get(value) ?? Types.get(value?.constructor);
            if (!type) {
                throw new TypeError(`Attribute "${key2}" is not valid type, must be either: String, Number, Boolean, null, undefined`);
            }
            types.set(key2, type);
            map.set(key2.toLowerCase(), key2);
        }
        const observedAttributes = [
            ...map.keys()
        ];
        return class extends superclass {
            static get observedAttributes() {
                return observedAttributes.concat(super.observedAttributes ?? []);
            }
            #data;
            constructor(...args){
                super(...args);
                this.#data = applyProps(this, new attrs1());
            }
            attributeChangedCallback(name, oldValue, newValue) {
                let key;
                if (key = map.get(name)) {
                    const prev = this.#data[key];
                    const next = types.get(key)(newValue);
                    if (!Object.is(prev, next)) {
                        this.#data[key] = next;
                        this.propertyChangedCallback?.(key, prev, next);
                    }
                }
                super.attributeChangedCallback?.(name, oldValue, newValue);
            }
        };
    }
;

const Listener = (body)=>new Function('event', `with(this){let fn=${body};return typeof fn=='function'?fn.call(this,event):fn}`)
;
const events = ()=>(superclass)=>superclass.prototype.dispatch
         ? superclass : class extends superclass {
            dispatchEvent(event) {
                const onEvent = `on${event.type}`;
                let pass = true;
                if (!(onEvent in this.constructor.prototype)) {
                    let fn = this[onEvent];
                    if (!fn) fn = Listener(this.getAttribute(onEvent));
                    pass = fn.call(this, event);
                }
                if (pass !== false) super.dispatchEvent(event);
                return pass;
            }
            dispatch(name, detail, init) {
                return this.dispatchEvent(new CustomEvent(name, {
                    detail,
                    ...init
                }));
            }
        }
;

const lifecycle = ()=>(superclass)=>class extends mixter(superclass, events()) {
            isMounted = false;
            connectedCallback() {
                super.connectedCallback?.();
                if (!this.isMounted) {
                    this.isMounted = true;
                    this.dispatch('mounted');
                }
            }
            disconnectedCallback() {
                super.disconnectedCallback?.();
                queueMicrotask(()=>{
                    if (!this.isConnected) {
                        this.isMounted = false;
                        this.dispatch('unmounted');
                    }
                });
            }
        }
;

const props = (props1)=>(superclass)=>class extends superclass {
            constructor(...args){
                super(...args);
                Object.defineProperties(this, Object.fromEntries(Object.entries(new props1()).map(([key, value])=>[
                        key,
                        {
                            configurable: true,
                            enumerable: true,
                            value
                        },
                    ]
                )));
            }
        }
;

const shadow = (init = '', html = '')=>(superclass)=>class extends superclass {
            root = (()=>{
                const root = this.attachShadow(typeof init === 'object' ? init : {
                    mode: 'open'
                });
                root.innerHTML = typeof init === 'string' ? init : html;
                return root;
            })();
        }
;

const argsOf = (x1)=>{
    const d = x1.indexOf('{') + 1;
    if (!d || x1.slice(0, d).includes(')')) return [];
    x1 = x1.slice(d);
    let match;
    do {
        match = false;
        x1 = x1.replace(/`((\\`)?[^`])+?`|'((\\')?[^'])+?'|"((\\")?[^"])+?"|\([^{[(]*?\)|\{[^{[(]*?\}|\[[^{[(]*?\]/g, ()=>{
            match = true;
            return '';
        });
    }while (match)
    return x1.split('}')[0].split(',').map((x)=>x.split(/[^\w\s$]+/g)[0].trim()
    );
};
 const argtor = (fn, keys = new Set())=>(argsOf('' + fn).forEach((x)=>!x || keys.add(x)
    ), keys)
;

const create = (target)=>{
    const NO_DEPS = Symbol();
    const REDUCER = Symbol();
    const entries = Object.entries(target);
    const mem = Object.fromEntries(entries);
    const register = (f)=>{
        f.keys = argtor(f.fn);
        f.keys.forEach((key)=>{
            const fx = effects.get(key);
            if (!fx) {
                console.warn('No effects for key:', key);
                return;
            }
            fx.push(f);
        });
        !f.keys.size && effects.get(NO_DEPS).push(f);
        f.values = {};
        return run(f);
    };
    const update = (f)=>{
        f.pass = !f.keys.size;
        for (const key of f.keys){
            const value = target[key];
            if (value == null) {
                f.pass = false;
                return;
            }
            if (!Object.is(f.values[key], value)) {
                f.values[key] = value;
                f.pass = true;
            }
        }
    };
    const run = (f)=>{
        update(f);
        f.dispose?.();
        if (!f.pass) return;
        const result = f.fn(f.values);
        if (f.cb?.(result) === false) return;
        if (f.target != null) {
            target[f.target] = result;
            return true;
        } else {
            if (typeof result === 'function') {
                f.dispose = ()=>{
                    result();
                    f.dispose = null;
                };
            }
        }
    };
    const context = {
        $: {}
    };
    const $ = context.$;
    $.$ = $;
    accessors(context.$, target, (key)=>({
            get: ()=>mem[key]
            ,
            set (v) {
                if (v === REDUCER) {
                    v = reducer.initial;
                    reducer.target = key;
                    if (register(reducer)) return;
                    if (v == null) return;
                }
                target[key] = v;
            }
        })
    );
    accessors(target, target, (key)=>({
            get: ()=>mem[key]
            ,
            set (v) {
                if (!Object.is(mem[key], v)) {
                    mem[key] = v;
                    effects.get(key).forEach(run);
                }
            }
        })
    );
    let reducer;
    $.reduce = (fn, initial)=>{
        reducer = {
            fn,
            initial
        };
        return REDUCER;
    };
    $.effect = (fn, cb)=>register({
            fn,
            cb
        })
    ;
    let effects;
    $.cleanup = ()=>{
        if (effects) {
            for (const fns of effects.values())fns.forEach((f)=>f.dispose?.()
            );
        }
        effects = new Map([
            [
                NO_DEPS,
                []
            ],
            ...entries.map(([key])=>[
                    key,
                    []
                ]
            ),
        ]);
    };
    $.cleanup();
    return $;
};
const state = (fn)=>(superclass)=>class extends mixter(superclass, lifecycle()) {
            host = this;
            constructor(...args){
                super(...args);
                const context = create(this);
                on()(this, 'mounted', ()=>fn.call(this, context)
                );
                on()(this, 'unmounted', context.cleanup);
            }
        }
;

function mixter(a, ...mixins) {
    return mixins.reduce((c, mixin)=>mixin(c)
    , a);
}

const observeNodes = (nodes, observer, options)=>nodes?.forEach((node)=>observer.observe(node, options)
    )
;
const nodesToText = (nodes)=>nodes?.map((node)=>{
        const text = node.textContent;
        return text.trim().length ? text : '';
    }).join('') ?? ''
;
const slotted = (slots)=>({
        get nodes () {
            return slots.map((slot)=>slot.assignedNodes()
            ).flat(Infinity);
        },
        get elements () {
            return slots.map((slot)=>slot.assignedElements()
            ).flat(Infinity);
        },
        get firstChild () {
            return this.elements[0];
        }
    })
;
const onSlotChange = (el1, cb, fn = (el)=>[
        ...el.querySelectorAll('slot')
    ]
)=>on()(el1, 'slotchange', ()=>cb(slotted(fn(el1)))
    )
;
const onTextChange = (el, cb, fn)=>{
    let observer;
    const off = onSlotChange(el, ({ nodes  })=>{
        observer?.disconnect();
        if (nodes.length) {
            observer = new MutationObserver(()=>cb(nodesToText(nodes))
            );
            observeNodes(nodes, observer, {
                characterData: true
            });
            cb(nodesToText(nodes));
        }
    }, fn);
    return ()=>{
        observer?.disconnect();
        off();
    };
};

/**
 * Compiles a syntax definition.
 *
 * ```js
 * const r = await compile({
 *   foo: /[a-z]/,
 *   bar: /[0-9]/,
 * })
 * ```
 *
 * @param def The syntax definition to compile. Can be a promise returned by `import()`.
 */
const compile = async (def, keys = new Set()) => {
    // try to extract the definitions if we're given either module or promise
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    const R = (await ((await def?.default) ?? def))?.default ?? def;
    if (!R)
        throw new ReferenceError('Could not find definition for language');
    const map = {};
    const parts = [];
    // eslint-disable-next-line prefer-const
    for (let [name, value] of Object.entries(R)) {
        keys.add(name);
        if (Array.isArray(value)) {
            map[name] = await compile(value[1], keys);
            value = value[0];
        }
        parts.push(`(?<${name}>${value?.source ?? value})`);
    }
    const regexp = RegExp(parts.join('|'), 'gm');
    regexp.map = map;
    regexp.keys = keys;
    return regexp;
};
/**
 * Syntax highlights a string as html with the given syntax.
 *
 * ```js
 * const regexp = await compile({
 *   foo: /[a-z]+/,
 *   bar: [
 *     /[0-9]+/,
 *     {
 *       bar: /[0-5]+/,
 *     },
 *   ],
 * })
 * const html = syntax(regexp, 'hello 123 789 world')
 * ```
 *
 * @param regexp The syntax definition returned by `compile()`
 * @param s The string to highlight.
 */
const syntax = (regexp, s) => s.replace(RegExp('(' + regexp.source + ')|(?<catchbad>[&<>"\'])', 'gm'), (...args) => {
    const groups = args.pop();
    const entries = Object.entries(groups).filter(e => e[1] != null);
    if (!entries.length)
        return '';
    const [tag, value] = entries[0];
    if (tag === 'catchbad')
        return escape(value);
    if (regexp.map) {
        const sub = regexp.map[tag];
        if (sub)
            return `<span class="${tag}-outer">${syntax(sub, value)}</span>`;
    }
    return `<span class="${tag}">${escape(value)}</span>`;
});

const languages = {};
const compiled = new Map();
const style = /*css*/ `
:host {
  display: block;
}
* {
  color: var(--color);
}
*::selection {
  background: var(--selection) !important;
}
pre {
  margin: 0;
}
slot {
  display: none;
}`;
const keyStyle = (x) => /*css*/ `
.${x} {
  color: var(--${x});
  font-style: var(--${x}-style);
  font-weight: var(--${x}-weight);
}`;
class CodeSyntaxElement extends mixter(HTMLElement, shadow(/*html*/ `<style>${style}</style><slot></slot><pre><code></code></pre>`), attrs(class {
    codeHTML = String;
    language = String;
    theme = String;
}), props(class {
    syntax;
    code;
    codeRawText;
    highlight;
    onTextContent;
}), state(({ $, effect, reduce }) => {
    $.code = reduce(({ root }) => root.querySelector('code'));
    // prevents unstyled content from showing until attributes
    // have arrived and the actual syntax has compiled
    setTimeout(() => {
        if ($.highlight == null)
            $.highlight = escape;
    }, 100);
    effect(({ language }) => {
        if (language in languages)
            $.syntax = languages[language];
    });
    effect(async ({ root, syntax: syntax$1 }) => {
        let { style, regexp } = compiled.get(syntax$1) ?? {};
        if (!regexp || !style) {
            regexp = await compile(syntax$1);
            style = new CSSStyleSheet();
            await style.replace([...regexp.keys].map(keyStyle).join(''));
            compiled.set(syntax$1, { style, regexp });
        }
        root.adoptedStyleSheets = [style];
        $.highlight = (s) => syntax(regexp, s);
    });
    effect(({ root }) => onTextChange(root, text => {
        $.codeRawText = text;
    }));
    $.codeHTML = reduce(({ codeRawText, highlight }) => highlight(codeRawText));
    effect(({ code, codeHTML }) => {
        code.innerHTML = codeHTML;
    });
})) {
}

const patchElements = (elements, { theme = 'deep', getLanguage = el => [...el.classList]
    .filter(x => x.startsWith('language-'))[0]
    .split('-')
    .pop(), } = {}) => elements.forEach(oldEl => {
    const newEl = new CodeSyntaxElement();
    newEl.textContent = oldEl.textContent;
    newEl.setAttribute('theme', theme);
    newEl.setAttribute('language', getLanguage(oldEl));
    oldEl.replaceWith(newEl);
});
const patchPreCodeElements = (opts) => patchElements([...document.querySelectorAll('code')].filter(el => el.parentNode?.nodeName === 'PRE'), opts);

export { CodeSyntaxElement, compile, languages, patchElements, patchPreCodeElements, syntax };
//# sourceMappingURL=code-syntax.js.map
