import { escape } from 'html-escaper';
import { attrs, mixter, onTextChange, props, shadow, state } from 'mixter';
import { compile, syntax as syntaxHighlight } from './syntax';
export const languages = {};
const compiled = new Map();
const style = /*css*/ `
:host {
  display: block;
}
* {
  color: var(--color);
}
*::selection {
  background: var(--selection) !important;
}
pre {
  margin: 0;
}
slot {
  display: none;
}`;
const keyStyle = (x)=>/*css*/ `
.${x} {
  color: var(--${x});
  font-style: var(--${x}-style);
  font-weight: var(--${x}-weight);
}`
;
export class CodeSyntaxElement extends mixter(HTMLElement, shadow(/*html*/ `<style>${style}</style><slot></slot><pre><code></code></pre>`), attrs(class {
    codeHTML = String;
    language = String;
    theme = String;
}), props(class {
    syntax;
    code;
    codeRawText;
    highlight;
    onTextContent;
}), state(({ $ , effect , reduce  })=>{
    $.code = reduce(({ root  })=>root.querySelector('code')
    );
    // prevents unstyled content from showing until attributes
    // have arrived and the actual syntax has compiled
    setTimeout(()=>{
        if ($.highlight == null) $.highlight = escape;
    }, 100);
    effect(({ language  })=>{
        if (language in languages) $.syntax = languages[language];
    });
    effect(async ({ root , syntax  })=>{
        let { style: style1 , regexp  } = compiled.get(syntax) ?? {};
        if (!regexp || !style1) {
            regexp = await compile(syntax);
            style1 = new CSSStyleSheet();
            await style1.replace([
                ...regexp.keys
            ].map(keyStyle).join(''));
            compiled.set(syntax, {
                style: style1,
                regexp
            });
        }
        root.adoptedStyleSheets = [
            style1
        ];
        $.highlight = (s)=>syntaxHighlight(regexp, s)
        ;
    });
    effect(({ root  })=>onTextChange(root, (text)=>{
            $.codeRawText = text;
        })
    );
    $.codeHTML = reduce(({ codeRawText , highlight  })=>highlight(codeRawText)
    );
    effect(({ code , codeHTML  })=>{
        code.innerHTML = codeHTML;
    });
})) {
}
